import contextlib
import pickle
import logging

from django.core.management import BaseCommand

from telegram_bot import bot
from telegram_bot import conf, redis_conn


class Command(BaseCommand):
    help = 'Start telegram bot'

    def handle(self, *args, **options):
        """entrypoint"""

        def event_handler(message: dict) -> None:
            """message handler"""

            # not that key exp
            if message['data'].decode('utf-8') != conf['REDIS_EXP_KEY']:
                return

            # no messages to send
            if not (length := redis_conn.llen(conf['REDIS_MESSAGES_KEY'])):
                return

            messages = redis_conn.lrange(conf['REDIS_MESSAGES_KEY'], 0, length-1)
            redis_conn.ltrim(conf['REDIS_MESSAGES_KEY'], length, -1)

            for message in messages:
                message = pickle.loads(message)
                bot.send_raw(**message)

            redis_conn.delete(conf['REDIS_EXP_KEY'])

        # add redis psubscribe and run in thread
        pubsub = redis_conn.pubsub()
        redis_conn.config_set('notify-keyspace-events', 'Ex')
        pubsub.psubscribe(**{"__keyevent@0__:expired": event_handler})
        pubsub.run_in_thread(sleep_time=conf['REDIS_EXP_TIME'])
        logging.info('Running worker redis subscriber')

        # start bot
        with contextlib.suppress(KeyboardInterrupt, SystemExit):
            bot.start_polling()



from importlib import import_module

from django.apps import apps

from telegram_bot import conf


def autodiscover_tg_routers():
    """Automatic search and initialization of 'tg_router' functions (by default)"""
    for app_config in apps.get_app_configs():
        try:
            # Name of the module containing 'tg_router', e.g. 'myapp.tg_router'
            routers_module_name = f"{app_config.name}.{conf['MODULE_NAME']}"
            import_module(routers_module_name)

        except ImportError:  # tg_router not exists in that app
            pass



from typing import Tuple, Any, Optional, Union, Iterable

from django.core.checks import CheckMessage, Error

from telegram_bot import conf


def check_settings(**kwargs: Any) -> Tuple[CheckMessage]:
    return (  # noqa
        *check_int('REDIS_EXP_TIME', 1, _min=0),
        *check_str('REDIS_EXP_KEY', 2),
        *check_str('REDIS_MESSAGES_KEY', 3),
        *check_str('TOKEN', 4),
        *check_str('REDIS_URL', 5),
        *check_str('MODULE_NAME', 6)
    )


def check_int(key: str, error: int, _min: Optional[int] = None, _max: Optional[int] = None) -> Tuple[Error]:
    """
    Check if value in settings is integer and in range.

    :param key: key in settings
    :param error: error code
    :param _min: minimum value
    :param _max: maximum value
    :return: tuple of errors
    """
    error_msg = '{key} should be an integer{greater}{less}.'.format(
        key=key,
        greater=f', greater than {_min}' if _min is not None else '',
        less=f', less than {_max}' if _max is not None else ''
    )

    value = conf.get(key)
    errors = False
    if isinstance(value, int):
        if _min is not None and value < _min:
            errors = True
        if _max is not None and value > _max:
            errors = True

    if not errors:
        return tuple()
    return (Error(error_msg, id=f'telegram_bot.E{str(error).zfill(3)}'),)


def check_str(key: str, error: int,
              _not_equal: Optional[Union[str, Iterable[str]]] = None,
              _equal: Optional[Union[str, Iterable[str]]] = None) -> Tuple[Error]:
    """
    Check if value in settings is string and not empty.

    :param key: key in settings
    :param error: error code
    :param _not_equal: value to compare (not equal)
    :param _equal: value to compare (equal)
    :return: tuple of errors
    """
    def prepare_if_string(value: Union[str, Iterable[str]]) -> Iterable[str]:
        """Prepare value if it is string."""
        if isinstance(value, str):
            return (value,)
        return value

    error_msg = '{key} should be {not_equal}{equal}.'.format(
        key=key,
        not_equal=f'a string, not not_equal to {_not_equal}' if _not_equal else 'an empty string ',
        equal=f'a string, equal to {_equal}' if _equal else 'an empty string '
    )

    value = conf.get(key)
    errors = False
    if isinstance(value, str):
        if _not_equal is not None and value in prepare_if_string(_not_equal):
            errors = True
        elif _equal is not None and value not in prepare_if_string(_equal):
            errors = True

    if not errors:
        return tuple()
    return (Error(error_msg, id=f'telegram_bot.E{str(error).zfill(3)}'),)



from typing import TypedDict, Callable

from django.conf import settings

from telegram_bot.defaults import DEFAULTS


class Settings(TypedDict):
    REDIS_EXP_TIME: int
    REDIS_EXP_KEY: str
    REDIS_MESSAGES_KEY: str
    TOKEN: str
    REDIS_URL: str
    MODULE_NAME: str
    DEFAULT_KWARGS: Callable


# noinspection PyTypeChecker
conf: Settings = {**DEFAULTS, **getattr(settings, 'TELEGRAM_BOT', {})}



from django.apps import AppConfig
from django.core.checks import register, Tags

from telegram_bot.checks import check_settings


class TelegramBotAppConfig(AppConfig):
    name = "telegram_bot"
    verbose_name = "django-redis-aiogram"

    def ready(self) -> None:
        from telegram_bot.signals import autodiscover_tg_routers
        register(Tags.security)(check_settings)
        autodiscover_tg_routers()



import asyncio
import logging
import pickle
from asyncio import AbstractEventLoop
from dataclasses import dataclass, field
from typing import Dict, Any

from aiogram import Bot, Dispatcher, Router, exceptions
from aiogram.dispatcher.event.handler import CallbackType
from redis.client import Redis

from telegram_bot.redis import redis_conn
from telegram_bot.settings import conf


@dataclass(repr=False)
class TelegramBot:
    loop: AbstractEventLoop = field(default_factory=asyncio.new_event_loop)

    redis_conn: Redis = field(init=False, default=redis_conn)
    bot: Bot = field(init=False)
    dispatcher: Dispatcher = field(init=False)
    _router: Router = field(init=False, default_factory=Router)

    def __post_init__(self) -> None:
        async def setup() -> None:
            """setup bot and dispatcher"""
            self.bot = Bot(token=conf['TOKEN'])
            self.dispatcher = Dispatcher()

        self.loop.run_until_complete(setup())

    def start_polling(self) -> None:
        """start telegram polling"""
        self.dispatcher.include_router(self._router)
        self.loop.run_until_complete(self.dispatcher.start_polling(self.bot))

    def send_raw(self, function: str = 'send_message', **kwargs) -> None:
        """sending message"""

        async def send():
            max_retries = 20
            while max_retries:
                try:
                    await getattr(self.bot, function)(**kwargs)
                    return logging.info(log_text.format('message sent'))
                except exceptions.TelegramRetryAfter as e:
                    logging.exception(log_text.format(e))
                    max_retries -= 1
                    await asyncio.sleep(e.retry_after)
                except Exception as e:
                    logging.exception(log_text.format(e))
                    break

        default_kwargs = conf['DEFAULT_KWARGS'](function)
        kwargs = {**default_kwargs, **kwargs}
        log_text = 'send_message: {}'

        if self.loop.is_running():
            self.loop.create_task(send())
        else:
            self.loop.run_until_complete(send())

    @staticmethod
    def send_redis(function: str = 'send_message', **kwargs: Dict[str, Any]) -> None:
        """sending message via redis"""
        redis_conn.rpush(conf['REDIS_MESSAGES_KEY'], pickle.dumps({'function': function, **kwargs}))
        redis_conn.set(conf['REDIS_EXP_KEY'], 'EX', conf['REDIS_EXP_TIME'])

    def message(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'message' observer"""
        return self._add_router(event_name='message', *args, **kwargs)

    def edited_message(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'edited_message' observer"""
        return self._add_router(event_name='edited_message', *args, **kwargs)

    def channel_post(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'channel_post' observer"""
        return self._add_router(event_name='channel_post', *args, **kwargs)

    def edited_channel_post(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'edited_channel_post' observer"""
        return self._add_router(event_name='edited_channel_post', *args, **kwargs)

    def inline_query(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'inline_query' observer"""
        return self._add_router(event_name='inline_query', *args, **kwargs)

    def chosen_inline_result(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'chosen_inline_result' observer"""
        return self._add_router(event_name='chosen_inline_result', *args, **kwargs)

    def callback_query(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'callback_query' observer"""
        return self._add_router(event_name='callback_query', *args, **kwargs)

    def shipping_query(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'shipping_query' observer"""
        return self._add_router(event_name='shipping_query', *args, **kwargs)

    def pre_checkout_query(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'pre_checkout_query' observer"""
        return self._add_router(event_name='pre_checkout_query', *args, **kwargs)

    def poll(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'poll' observer"""
        return self._add_router(event_name='poll', *args, **kwargs)

    def poll_answer(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'poll_answer' observer"""
        return self._add_router(event_name='poll_answer', *args, **kwargs)

    def my_chat_member(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'my_chat_member' observer"""
        return self._add_router(event_name='my_chat_member', *args, **kwargs)

    def chat_member(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'chat_member' observer"""
        return self._add_router(event_name='chat_member', *args, **kwargs)

    def chat_join_request(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'chat_join_request' observer"""
        return self._add_router(event_name='chat_join_request', *args, **kwargs)

    def error(self, *args, **kwargs) -> CallbackType:
        """Decorator for 'error' observer"""
        return self._add_router(event_name='error', *args, **kwargs)

    def _add_router(self, *args, event_name: str, **kwargs) -> CallbackType:
        def wrapper(callback: CallbackType) -> CallbackType:
            observer = self._router.observers[event_name]
            observer.register(callback, *args, **kwargs)
            return callback

        return wrapper



#!/usr/bin/env python

from io import open
from setuptools import setup, find_packages

NAME = 'django-redis-aiogram'
VERSION = '1.0.6'


def read_md(file_path):
    with open(file_path) as fp:
        return fp.read()


setup(
    name=NAME,
    version=VERSION,
    install_requires=(
        'django>=4',
        'redis>=4.6',
        'aiogram>=3.0.0'
    ),
    author='Oleksii Kolosiuk',
    author_email='kolosyuk1@gmail.com',
    license='MIT',
    url='https://github.com/CorneiZeR/django-redis-aiogram',
    keywords='django redis aiogram docker',
    description='Running aiogram in neighbor container, sending messages to telegram via redis',
    long_description=read_md('README.md'),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    python_requires='>=3.8',
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python',
        "Programming Language :: Python :: 3",
        'Framework :: Django',
        'Framework :: Django :: 4',
    ]
)


from typing import Dict, Union, Callable

DEFAULTS: Dict[str, Union[int, str, Callable]] = {
    # event expiration time in redis
    'REDIS_EXP_TIME': 5,
    # redis key for handling expired event
    'REDIS_EXP_KEY': 'TELEGRAM_BOT_EXP',
    # redis key for collecting messages
    'REDIS_MESSAGES_KEY': 'TELEGRAM_BOT_MESSAGE',
    # name of the module to find
    'MODULE_NAME': 'tg_router',
    # default kwargs for telegram bot
    'DEFAULT_KWARGS': lambda _: {},
    # telegram bot token
    'TOKEN': '',
    # url for redis connection
    'REDIS_URL': ''
}
